
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <malloc.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <net/if.h>
#include <linux/if_tun.h>

#include "endian.h"
#include "pkt.h"

#define RCV_BUF_SIZE	65536
#define DEFAULT_DEVNAME	"/dev/net/tun"

#define	VERSION		"1.0"
#define NUMDUPEBITS	14
#define NUMDUPES	(1<<NUMDUPEBITS)

static int 	maxchilds=40,
		childs,
		dupetime=60;

struct dstdupe {
	u32	dst;
	time_t	time;
};

struct dstdupe	dupes[NUMDUPES];
int		dtail,
		dhead;

/*
 * exec on a child and wait for its return on demand
 */
static void run(int wait, const char *fmt, ...) {
	int	pid,
		status;

	pid=fork();
	if (!pid) {
		va_list va;
		va_start(va, fmt);
		execv(fmt, va);
		va_end(va);
	}

	if (pid<0) {
		printf("Fork failed\n");
		exit(1);
	}

	if (wait)
		waitpid(pid, &status, 0);
	else
		childs++;
}

/*
 * Convert an network order ipv4 address to string 
 * FIXME is this endian clean ?
 */
static void makeaddr(char *b, u32 addr) {
	sprintf(b, "%d.%d.%d.%d",
		(u8) addr&0xff,
		(u8) (addr>>8)&0xff,
		(u8) (addr>>16)&0xff,
		(u8) (addr>>24)&0xff);
}


/*
 * Insert dst into dupe detection
 */
static void insertdupe(u32 dst) {
	/* Insert into dupe array */
	dupes[dhead].time=time(NULL);
	dupes[dhead].dst=dst;
	/* Increase pointer and mask to NUMDUPE (2^n) */
	dhead++;
	dhead&=NUMDUPES-1;
}

/* 
 * Do a garbage collection on the dupe list
 */
static void deldupes(void ) {
	time_t		mintime=time(NULL)-dupetime;

	while(dtail!=dhead) {
		/* We found a dupe which is less than dupetime old */
		if (dupes[dtail].time>mintime)
			break;

		dtail++;

		/* Mask <NUMDUPES - Numdupes has to be 2^n */
		dtail&=(NUMDUPES-1);
	}
}

/*
 * Check for the existance of dst in the dupe container 
 */
static int checkdupe(u32 dst) {
	int	dt=dtail;
	
	while(dt!=dhead) {

		/* We found a dupe which is less than dupetime old */
		if (dupes[dt].dst == dst)
			return 1;

		dt++;
		/* Mask <NUMDUPES - Numdupes has to be 2^n */
		dt&=(NUMDUPES-1);
	}

	return 0;
}

/*
 * Process a TUN packet
 * - Check for dupe and drop if duplicate destination
 * - Fork the dial script in case of no dupe and insert into dupe detection
 */
static int process_pkt(char *buf, int len, int maxchilds, char *pscript) {
	char		src[16],
			dst[16];
	struct tunpkt	*tp=(struct tunpkt *) buf;

	/* We only care about version 4 now */
	if (tp->ip.version != 4)
		return 0;

	makeaddr(src, tp->ip.saddr);		
	makeaddr(dst, tp->ip.daddr);		

	if (checkdupe(tp->ip.daddr)) {
		printf("Dropping packet from %s to %s - dupe\n",
				src, dst);
		return 0;
	}

	/* printf("Received packet from %s to %s\n", src, dst); */
	if (childs > maxchilds)
		printf("Had to drop packet due to maxchilds %d from %s to %s\n", 
				maxchilds, src, dst);
	else {
		printf("Running child %s for %s to %s\n", pscript, src, dst);
		run(0, pscript, "diald child", src, dst, NULL);
		insertdupe(tp->ip.daddr);
	}
	return 1;
}

/*
 * Wait for input on the TUN device - reap children and delete dupes
 */
static void process_io(int dev, int maxchilds, char *pscript) {
	char	*pktbuf;
	int	l,
		status;
	fd_set	fds;
	struct	timeval tv;

	if ((pktbuf=malloc(RCV_BUF_SIZE)) == 0) {
		printf("Malloc failed\n");
		close(dev);
		exit(1);
	}

	while(1) {
		FD_ZERO(&fds);
		FD_SET(dev, &fds);

		/* Return every second to do housekeeping */
		tv.tv_sec=1;
		tv.tv_usec=0;

		select(dev+1, &fds, NULL, NULL, &tv);

		/* Read packet if available */
		if (FD_ISSET(dev, &fds)) {
			l=read(dev, pktbuf, RCV_BUF_SIZE);
			if (l > 0) 
				process_pkt(pktbuf, l, maxchilds, pscript);
		}

		/* Housekeeping - Reap children and delete dupes */
		while(waitpid(-1, &status, WNOHANG) > 0)
			childs--;

		deldupes();
	}	
}

/* 
 * Alloc a TUN device and return its name and FD
 */
static int tun_alloc(char *devname, char *ifname) {
	struct ifreq 	ifr;
	int		fd, err;

	if((fd=open(devname, O_RDWR)) < 0 )  {
		printf("Could not open %s\n", devname);
		exit(1);
	}

	memset(&ifr, 0, sizeof(ifr));

	/* Flags: IFF_TUN   - TUN device (no Ethernet headers) 
	*        IFF_TAP   - TAP device  
	*
	*        IFF_NO_PI - Do not provide packet information  
	*/ 
	ifr.ifr_flags = IFF_TUN; 
	/* If ifname is filled */
	if(*ifname)
		strncpy(ifr.ifr_name, ifname, IFNAMSIZ);

	if((err=ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
		close(fd);
		printf("Could not create interface %s\n", ifname);
		exit(1);
	}

	ioctl(fd, TUNSETNOCSUM, 1);
	
	/* Copy back ifname we got */
	strcpy(ifname, ifr.ifr_name);
	return fd;
}              


static void help(int ret) {
	printf("diald %s Copyright 2003 Florian Lohoff <flo@rfc822.org>\n\n", VERSION);
	printf("-s [startscript]	- script to be run with tun iface created\n");
	printf("-p [packetscript]	- script to be run on every successfull packet\n");
	printf("-d <devicename>		- TUN device name (Default: /dev/net/tun)\n");
	printf("-i <ifname>		- interface name (Default: selected automatically)\n");
	printf("-m <maxchilds>		- max # of children (Default: 40)\n");
	printf("-u <dupetime>		- dupe seconds (Default: 60)\n");
	printf("-h 			- this help\n\n");
	exit(ret);
}

int main(int argc, char **argv) {
	int	dev,
		ret;
	char	*devname=DEFAULT_DEVNAME,
		*sscript=NULL,
		*pscript=NULL;
	char	ifname[IFNAMSIZ+1];

	/* Mark as empty */
	ifname[0]=0x0;

	while((ret = getopt(argc,argv,"hd:i:s:m:p:u:")) != -1) {
		switch(ret) {
			case 'd':
				devname=strdup(optarg);
				break;
			case 's':
				sscript=strdup(optarg);
				break;
			case 'i':
				strncpy(ifname, optarg, IFNAMSIZ);
				break;
			case 'm':
				maxchilds=atoi(optarg);
				break;
			case 'u':
				dupetime=atoi(optarg);
				break;
			case 'p':
				pscript=strdup(optarg);
				break;
			case 'h':
				help(0);
				break;
		}
	}

	if (!pscript || !sscript)
		help(1);

	/* FIXME Check for required options */
	dev=tun_alloc(devname, ifname);	

	if (!dev)
		exit(1);

	printf("Created tun device %s\n", ifname);

	if(sscript) {
		printf("Running start script %s\n", sscript);
		/* Execute start script (Setting up the tun interface) and
		 * wait for its excution end */
		run(1, sscript, "diald start", ifname, NULL);
	}

	printf("Waiting for packets\n");
	process_io(dev, maxchilds, pscript);

	exit(0);
}
